from tornado.options import options, define
from .json_utils import dumps
import asyncio, aioredis
import logging
import redis

async def connect():
    loop = asyncio.get_event_loop()
    redis = await aioredis.create_redis(options.redis_url, loop=loop)
    logging.info("connect to redis: %s", options.redis_url)
    return redis


async def pool():
    loop = asyncio.get_event_loop()
    pool = await aioredis.create_pool(options.redis_url,
                    minsize=5, maxsize=10, loop=loop)
    logging.info("pool to redis: %s", options.redis_url)
    return pool



from tornado.ioloop import IOLoop
from tornado.options import options
from tornado.platform.asyncio import AnyThreadEventLoopPolicy
from concurrent.futures import ThreadPoolExecutor
from tornado import autoreload
from ..web.settings import settings
from ..utils.rd_session import connect
from ..utils.json_utils import dumps, loads
from ..utils.config import extend
from . import perform
import importlib
import threading
import asyncio
import logging

LOGGER = logging.getLogger(__name__)

'''
    Main worker file. It subscribes to the
    redis queue and broadcasts activity
'''

class Worker:

    def __init__(self, ioloop):
        self.ioloop = ioloop
        self.ioloop.set_default_executor(ThreadPoolExecutor())
        self.procedures = importlib.import_module(options.procedures)


    def get_params(self, content):
        if isinstance(content["params"], dict):
            return [],content["params"]
        elif isinstance(content["params"], list):
            return content["params"],{}
        else:
            raise Exception("Params neither list or dict")


    async def perform(self, content):
        actor = content.get("current_user")
        args, kwargs = self.get_params(content)
        method = content["method"]
        content["thread"] = threading.get_ident()
        if ":" in method:
            mod_name,func_name = method.split(":")
            procedure = getattr(importlib.import_module(mod_name), func_name)
        else:
            procedure = getattr(self.procedures, method)
        try:
            LOGGER.info("%r %s %s", method, args, kwargs)
            result = await perform.perform(actor, procedure, args, kwargs)
            content["result"] = result
            if isinstance(result, perform.Context):
                content["is_context"] = True
        except Exception as ex:
            content["error"] = {
                "code": -32000,
                "message": str(ex)
            }
        return content


    async def _broadcast_(self, signal, message, filter_clients=None):
        msg = {
            "signal": signal,
            "message": message,
            "filter_clients": filter_clients
        }
        await self.broadcast_redis.publish("broadcast", dumps(msg))


    def broadcast(self, signal, message, filter_clients=None):
        '''
            a thread safe way to broadcast
        '''
        self.ioloop.spawn_callback(self._broadcast_,
            signal, message, filter_clients)


    async def do_work(self, val):
        content = loads(val[1].decode('utf-8'))
        if content.get("id"):
            await self._broadcast_('broadcast', {"working": content["id"]})
        content = await self.perform(content)
        if content.get("reply"):
            await self.broadcast_redis.publish(content["reply"], dumps(content))
            LOGGER.info("published reply %s", content)


    async def work(self):
        ''' Connects to redis a awaits messages from queue '''
        self.broadcast_redis = await connect()
        self.work_q = await connect()
        LOGGER.info("ready for work")
        while True:
            val = await self.work_q.blpop(options.redis_work_q)
            if val:
                await self.do_work(val)
        self.work_q.close()
        await self.work_q.wait_closed()
        self.broadcast_redis.close()
        await self.broadcast_redis.wait_closed()


    async def put(self, content):
        await self.broadcast_redis.execute('rpush',options.redis_work_q, dumps(content))


def main():
    ''' Creates Worker and starts tornado '''
    asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())
    values = settings()
    if options.settings_extend:
        values = extend(options.settings_extend, values)

    assert(options.redis_url)
    autoreload.start()
    ioloop = IOLoop.current()
    perform._manager_ = Worker(ioloop=ioloop)
    for _ in range(options.workers):
        ioloop.spawn_callback(perform._manager_.work)
    try:
        ioloop.start()
    except KeyboardInterrupt:
        logging.info("stopping")



from tornado.options import options
from tornado.ioloop import IOLoop
from tornado.web import Application
from ..work.manager import Manager
from ..utils.config import extend
from .settings import settings
from .urls import urls
import importlib
import logging

LOGGER = logging.getLogger(__name__)

'''
    Simple Tornado Server
'''


def make_app():
    values = settings()

    if options.settings_extend:
        values = extend(options.settings_extend, values)

    pcs = importlib.import_module(options.procedures)

    if options.redis_url:
        from ..work.redis_manger import RedisManager
        values["manager"] = RedisManager(pcs, options.workers)
    else:
        logging.info("local manager")
        values["manager"] = Manager(pcs, options.workers)

    paths = urls()

    if options.urls_extend:
        paths = extend(options.urls_extend, paths)

    return Application(paths, **values)


def main():
    app = make_app()
    app.listen(options.port)
    LOGGER.info("listening on port %s", options.port)

    ioloop = IOLoop.current()
    try:
        ioloop.start()
    except KeyboardInterrupt:
        logging.info("stopping")
        ioloop.stop()



from bluemax import context
import time
import asyncio
from tornado import gen


__all__ = ["add", "a_add", "t_add"]


def add(a=2, b=2, sleep_for=0.1):
    """ returns a plus b syncronous."""
    result = a+b
    time.sleep(sleep_for)
    context.broadcast_on_success("action", {
        "name": "add",
        "result": result,
        "user": context.get_current_user() })
    return "add", result


async def a_add(a=2, b=3, sleep_for=0.2):
    """ returns a plus b, async native."""
    result = a+b
    await asyncio.sleep(sleep_for)
    context.broadcast_on_success("action", {
        "name": "a_add",
        "result": result,
        "user": context.get_current_user() })
    return "a_add", result


@gen.coroutine
def t_add(a=2, b=4, sleep_for=0.3):
    """ returns a plus b, async tornado."""
    result = a+b
    yield gen.sleep(sleep_for)
    context.broadcast_on_success("action", {
        "name": "t_add",
        "result": result,
        "user": context.get_current_user() })
    return "t_add", result



from tornado.websocket import WebSocketHandler
from tornado.web import HTTPError
from tornado import gen
from .authentication import UserMixin
from ..utils import json_utils
import functools
import logging
import uuid

'''
    Implentation of the json rpc 2.0 spec.
'''

LOGGER = logging.getLogger(__name__)

class JsonRpcException(Exception):

    def __init__(self, code, message):
        super().__init__(self, message)
        self.js_code = code
        self.js_message = message


class JsonRpcHandler(UserMixin, WebSocketHandler):

    def initialize(self):
        self.id = str(uuid.uuid4())


    @property
    def manager(self):
        return self.settings['manager']


    @property
    def procedures(self):
        return self.manager.procedures


    def open(self):
        user = self.get_current_user()
        if user:
            self.manager.add_client(self)
            LOGGER.info("user: %s", user)
            self.write_message(json_utils.dumps({
                "signal": "set_user",
                "message": user
            }))
        elif self.settings["debug"] is False:
            raise HTTPError(403)
        else:
            self.manager.add_client(self)


    def get_params(self, content):
        if isinstance(content["params"], dict):
            return [],content["params"]
        elif isinstance(content["params"], list):
            return content["params"],{}
        else:
            raise JsonRpcException(-32602, "Params neither list or dict")


    @gen.coroutine
    def perform(self, content, result):
        method = content["method"]
        if method == "authenticate":
            result["result"] = self.get_transfer_user(content["params"][0])
            self.write_message(json_utils.dumps(result))
        else:
            content["client_id"] = self.id
            content["current_user"] = self.get_current_user()
            future = yield self.manager.put(content)
            if future:
                future.add_done_callback(functools.partial(self.work_done, result))


    @gen.coroutine
    def on_message(self, data):
        content = json_utils.loads(data)
        LOGGER.info(content)
        result = {
            "jsonrpc": "2.0",
        }
        try:
            if content.get("jsonrpc") != '2.0':
                raise JsonRpcException(-32600, "protocol not supported")
            ref = content.get("id")
            if ref:
                result["id"] = ref

            method = content.get('method')
            if method is None:
                raise JsonRpcException(-32600, "no method")
            if method[0] == "_":
                raise JsonRpcException(-32600, "method private")
            if method in self.procedures.__all__ + ["authenticate"]:
                yield self.perform(content, result)
                result = None
            else:
                raise JsonRpcException(-32600, f"no such method: {method}")

            if ref is None:
                result = None # its a notification

        except JsonRpcException as ex:
            LOGGER.exception(ex)
            result["error"] = {
                "code": ex.js_code,
                "message": ex.js_message
            }
        except Exception as ex:
            LOGGER.exception(ex)
            result["error"] = {
                "code": -32000,
                "message": str(ex)
            }
        finally:
            if result:
                LOGGER.debug(repr(result))
                self.write_message(json_utils.dumps(result))


    def work_done(self, result, future):
        try:
            output = future.result()
            result["result"] = output
        except Exception as ex:
            result["error"] = {
                "code": -32000,
                "message": str(ex)
            }
        self.write_message(json_utils.dumps(result))


    def on_close(self):
        self.manager.remove_client(self)
